# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 3906 $
$LastChangedDate:: 2020-11-24 11:20:27 +0100#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
import asyncio
import logging
import websockets
import json_tricks

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)
logger = logging.getLogger('websockets.server')


class HerokuServer:
    """Server class for Model-As-A-Service at Heroku.

    This is the server part of a client-server websocket application.
    - Runs forever
    - Waits for calls from client
    - After data receive, executes model
    - Sends back results as json string
    - Sends back log. (Not implemented yet)
    """

    run_fnc = None

    def __init__(self, server_uri: str = "0.0.0.0", server_port: int = None):
        """
        Heroku-Server which will send and receive data from a client

        Parameters
        ----------
        server_uri: str
            URI of the server e.g. "symspace-demo-app10.herokuapp.com"
            On the Heroku-Dyno the URI is "0.0.0.0" (Default)
        server_port: int
            Listening port of the server.
            This will automatically set by the Heroku environment variable "PORT",
            but can also be set manually for debugging
        """
        self.server_uri = server_uri
        if server_port is None:
            self.server_port = os.environ.get('PORT')  # Heroku dynamically assigns a port
        else:
            self.server_port = server_port  # for debugging

    def start(self):
        logger.debug("Started server:\n{}/{}".format(self.server_uri, self.server_port))

        start_server = websockets.serve(self.server, self.server_uri, self.server_port)

        logger.info(f"Started event loop.\n---------------------------------------")
        loop = asyncio.get_event_loop()
        loop.run_until_complete(start_server)
        loop.run_forever()

    @staticmethod
    def set_run_fnc(fnc):
        """Workaround to set the core run function of the server externally"""
        HerokuServer.run_fnc = fnc

    @staticmethod
    async def server(websocket, path):
        """

        Parameters
        ----------
        websocket
            Websocket object
        path
        """
        '''wait for client to connect'''
        user_data = await websocket.recv()
        logger.info('Client connected:\n\t{}\n---------------------------------------'.format(user_data))

        '''wait for data from client'''
        logger.info("Waiting for data ...")
        input_data_json = await websocket.recv()
        logger.info("Receive data done")

        '''process data'''
        logger.info("Process data ...")
        input_data = json_tricks.loads(input_data_json)
        output_data = HerokuServer.run_fnc(input_data, logger)
        logger.info("Process data done")

        '''send data'''
        logger.info("Sending data ...")
        output_data_json = json_tricks.dumps(output_data)
        await websocket.send(output_data_json)
        await websocket.send("Currently no messages are logged!")  # Todo: implement sending of recent log messages
        logger.info("Send done")
