# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 5845 $
$LastChangedDate:: 2022-05-03 18:32:29 +0200#$

This file is part of SyMSpace.

Copyright (c) 2021, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import re
import shutil
import time
import json
import glob
import numpy as np
from openpyxl import load_workbook

from symspace.wrapper.symwrapper import SymProject

import componentspace.osutils.fileutils as fu
from symspace.osutils.pathutils import Paths
from componentspace.builders.scripts import build_external_material
from componentspace.material_utils.material_functions import GeneralMaterialFunctions
from componentspace.material_utils.batch_material import MaterialBatchComponent

from componentspace.material_component import MaterialComponent

try:
    from symspace.osutils.multiprocessing import ExternalMultiprocessing
except ImportError:
    raise Exception('Update SyMSpace to the latest version')


class MaterialBomatec(MaterialBatchComponent):
    """

    """

    def createDirectory(self, material):
        """
        Creates a directory with the name of the geometry and the component class
        Copies directories /rst, /help, and /doc

        Parameters
        ----------
        material : dict
            Dictionary of material properties

        """
        # Remove directory

        fname = material['FileName']
        path_material = os.path.join(self.path, fname)

        if os.path.exists(path_material):
            shutil.rmtree(path_material, onerror=fu.on_rm_error)
            # wait until OS is finished
            cnt = 0
            while os.path.exists(path_material):
                time.sleep(1)
                cnt = cnt + 1
                if cnt > 3:
                    raise ValueError('Cannot remove path!')

        # Create material directory
        if not os.path.exists(path_material):
            os.mkdir(path_material)

        shutil.copy(os.path.join(self.path, 'BomatecMagnets.mop'),
                    os.path.join(path_material, material['FileName'] + '.mop'))

        if material['gesintert']:
            mat_description = 'sintered ' + material['Werkstoff']
        else:
            mat_description = material['Werkstoff']
        # Copy rst
        fu.copyFileAndReplace(os.path.join(self.path, 'index.rst'), os.path.join(path_material, 'index.rst'),
                              [('__ComponentTitle__',
                                material['Bomatec Bezeichnung'] + '\n' + len(material['Bomatec Bezeichnung']) * '='),
                               ('__Material__', mat_description),
                               ('__DemagPic__', fname + '.png'),
                               ('__PDF__', fname + '.pdf')])

        fu.copyTree(os.path.join(self.path, 'rst'), os.path.join(path_material, 'rst'))

        # Copy src
        fu.copyTree(os.path.join(self.path, 'src'), os.path.join(path_material, 'src'))

        # Copy png file
        png_file = os.path.join(self.path, 'resources', 'PDF', fname + '.png')
        if os.path.isfile(png_file):
            shutil.copy(png_file, os.path.join(path_material, 'rst', fname + '.png'))

        # Copy pdf file
        pdf_file = os.path.join(self.path, 'resources', 'PDF', fname + '.pdf')
        if os.path.isfile(pdf_file):
            shutil.copy(pdf_file, os.path.join(path_material, 'rst', fname + '.pdf'))

        # Modify material file
        prj = SymProject.load(os.path.join(path_material, material['FileName'] + '.mop'))
        prj['Project.Material.Name'].setData(material['Bomatec Bezeichnung'])
        prj['Project.Material.rho'].setData(material['Dichte ca'])
        prj['Project.Material.E'].setData(material['modulus of elasticity E'])
        prj['Project.Material.Rm'].setData(material['flexural strength'])
        prj['Project.Material.HV'].setData(material['vickers hardness HV'])
        prj['Project.Material.Cm'].setData(material['specific heat capacity'])
        prj['Project.Material.lambda'].setData(material['thermal conductivity'])
        prj['Project.Material.rhoel'].setData(material['specific el. resistance'])
        prj['Project.Material.Br'].setData(material['Br min'])
        prj['Project.Material.mur'].setData(material['permeability 20°C µr'])
        prj['Project.Material.alpha_BrT'].setData(material['temperature coefficient Br'] / 100.)
        prj['Project.Material.Tmax'].setData(material['T max'])
        prj['Info.Name'].setData(material['Bomatec Bezeichnung'] + '_Bomatec')
        prj['Info.FunctionName'].setData(material['Bomatec Bezeichnung'])
        prj['Info.DisplayName'].setData(material['Bomatec Bezeichnung'])
        prj['Info.WebPath'].setData('Material.Metal.MagneticAlloy.{}'.format(material['Werkstoff']))

        # Read JSON file for demagnetization
        json_file = os.path.join(self.path, 'resources', 'Hc_lim', fname + '.json')
        if os.path.exists(json_file):
            with open(json_file) as fh:
                demag = json.loads(fh.read())

            Hcb = np.array([])
            for key in demag:
                theta = float(key.replace('degC', ''))
                Hcb = np.append(Hcb, np.array([theta, demag[key] * -1000]))

            Hcb = Hcb.reshape(len(demag.keys()), 2)
            idx = np.argsort(Hcb, axis=0)

            prj['Project.Material.HvsTheta.Theta'].setData(Hcb[idx[:, 0], 0])
            prj['Project.Material.HvsTheta.Hcb'].setData(Hcb[idx[:, 0], 1])

        prj.save()
        prj.close()

    def buildComponent(self, keep_mopc: bool = False):
        """
        """
        file_model = os.path.join(self.path, self.name + '.mop')
        if os.path.isfile(file_model):
            self.sym_mdl = SymProject.load(file_model)
        else:
            raise Exception('Model {} file does not exist!'.format(file_model))

        # Read Excel file
        excl_file = self.sym_mdl.getObject('Info.BomatecDatabase').getData()
        self.sym_mdl.close()

        wb = load_workbook(excl_file)

        try:
            ws = wb['Bomatec Qualität']
        except Exception:
            '''get_sheet_by_name is deprecated'''
            ws = wb.get_sheet_by_name('Bomatec Qualität')

        rows_iter = ws.iter_rows(min_col=ws.min_column, min_row=3, max_col=ws.max_column, max_row=3)
        title = [[cell.value for cell in row] for row in rows_iter][0]

        select = ('Bomatec Bezeichnung', 'Werkstoff', 'Temperaturstabil', 'Korrosionsstabil',
                  'isoptrop', 'anisotrop', 'gesintert', 'gegossen', 'Br min', 'Br nom', 'Dichte ca',
                  'modulus of elasticity E', 'flexural strength', 'vickers hardness HV', 'specific heat capacity',
                  'thermal conductivity', 'specific el. resistance', 'permeability 20°C µr',
                  'temperature coefficient Br', 'T max')

        lst_val = []
        for k in range(4, ws.max_row + 1):
            # for k in range(301, 350 + 1):

            rows_iter = ws.iter_rows(min_col=ws.min_column, min_row=k, max_col=ws.max_column, max_row=k)
            row = [[cell.value for cell in iter] for iter in rows_iter][0]
            if row[title.index('Bomatec Bezeichnung')] is None:
                continue

            val = {}
            for name in select:
                cell = row[title.index(name)]
                # print(name + ' :: ' + str(cell))
                if name == 'Bomatec Bezeichnung':
                    val[name] = cell
                    val['FileName'] = cell.replace('/', '_').replace(' ', '')
                elif isinstance(cell, str) and '~' in cell:
                    rng = cell.split('~')
                    val[name] = (float(rng[0].strip()) + float(rng[1].strip())) / 2
                elif isinstance(cell, str) and '-' in cell:
                    rng = cell.split('-')
                    try:
                        val[name] = (float(rng[0].strip()) + float(rng[1].strip())) / 2
                    except:
                        val[name] = None
                else:
                    val[name] = cell

            lst_val.append(val)

        n_mat = len(lst_val)

        xmp = ExternalMultiprocessing(max_cpu_percent=75.0, max_ram_percent=75.0, stdout=None)
        DebugMode = False

        for index, mat in enumerate(lst_val):
            print('Build Material #{:d} / {:d} '.format(index + 1, n_mat))
            self.createDirectory(mat)

            if DebugMode:
                '''single internal process -> slow only for debugging'''
                comp = MaterialComponent(os.path.join(self.path, mat['FileName'], mat['FileName'] + '.mop'))
                comp.base_directory = self.path
                comp.buildComponent()
                comp.sym_mdl.close()
            else:
                '''build with external processes for better/multicore cpu usage'''
                comp_file = os.path.join(self.path, mat['FileName'], mat['FileName'] + '.mop')
                base_directory = self.path
                script_py = build_external_material.__file__
                cmd_args = [Paths.getCSyMSpaceExe(), script_py, comp_file, base_directory]
                xmp.startProcess(cmd_args, wait_to_start=False, wait_to_finish=False)
                xmp.wait()

        '''wait until all processes are done'''
        xmp.waitForCompletion()

    def _createSingleCompfileFromTemplate(self, comp_file: str, material: dict):
        path_material = os.path.dirname(comp_file)
        component_name = material['component_name']
        pdf_name = material['pdf_name']
        png_name = material['png_name']

        # Create material directory
        if not os.path.exists(path_material):
            os.mkdir(path_material)

        # shutil.copy(os.path.join(self.path, 'BomatecMagnets.mop'),
        #             os.path.join(path_material, material['component_name'] + '.mop'))

        if material['gesintert']:
            mat_description = 'sintered ' + material['Werkstoff']
        else:
            mat_description = material['Werkstoff']
        # Copy rst
        fu.copyFileAndReplace(os.path.join(self.path, 'index.rst'), os.path.join(path_material, 'index.rst'),
                              [('__ComponentTitle__',
                                material['Bomatec Bezeichnung'] + '\n' + len(material['Bomatec Bezeichnung']) * '='),
                               ('__Material__', mat_description),
                               ('__DemagPic__', png_name + '.png'),
                               ('__PDF__', pdf_name + '.pdf')])

        fu.copyTree(os.path.join(self.path, 'rst'), os.path.join(path_material, 'rst'))

        ''' Copy src '''
        fu.copyTree(os.path.join(self.path, 'src'), os.path.join(path_material, 'src'))

        ''' Copy png file '''
        png_file = os.path.join(self.path, 'resources', 'PDF', png_name + '.png')
        if os.path.isfile(png_file):
            shutil.copy(png_file, os.path.join(path_material, 'rst', png_name + '.png'))

        ''' Copy pdf file '''
        pdf_file = os.path.join(self.path, 'resources', 'PDF', pdf_name + '.pdf')
        if os.path.isfile(pdf_file):
            shutil.copy(pdf_file, os.path.join(path_material, 'rst', pdf_name + '.pdf'))

        ''' Modify material file '''
        prj = SymProject.load(comp_file)
        # prj['Project.Material.Type'].setData('Ferromagnetic material')
        # prj['Project.Material.Name'].setData(component_name)

        GeneralMaterialFunctions.set_filter(prj, material)
        GeneralMaterialFunctions.set_info(prj, material)
        GeneralMaterialFunctions.set_general_properties(prj, material)
        GeneralMaterialFunctions.set_electromagnetic_properties(prj, material)
        GeneralMaterialFunctions.set_mechanic_properties(prj, material)
        GeneralMaterialFunctions.set_thermal_properties(prj, material)

        prj.save()

    def extractMaterialData(self, keep_mopc: bool = False):
        """
        """
        # Todo: implementation is not complete

        # Read Excel file
        excl_file = self.sym_mdl.getObject('Info.BomatecDatabase').getData()
        self.sym_mdl.close()

        wb = load_workbook(excl_file)

        try:
            ws = wb['Bomatec Qualität']
        except Exception:
            '''get_sheet_by_name is deprecated'''
            ws = wb.get_sheet_by_name('Bomatec Qualität')

        rows_iter = ws.iter_rows(min_col=ws.min_column, min_row=3, max_col=ws.max_column, max_row=3)
        title = [[cell.value for cell in row] for row in rows_iter][0]

        select = ('Bomatec Bezeichnung', 'Werkstoff', 'Temperaturstabil', 'Korrosionsstabil',
                  'isoptrop', 'anisotrop', 'gesintert', 'gegossen', 'Br min', 'Br nom', 'Dichte ca',
                  'modulus of elasticity E', 'flexural strength', 'vickers hardness HV', 'specific heat capacity',
                  'thermal conductivity', 'specific el. resistance', 'permeability 20°C µr',
                  'temperature coefficient Br', 'T max')

        all_materials_dictionary = {}
        unit_iter = ws.iter_rows(min_col=ws.min_column, min_row=2, max_col=ws.max_column, max_row=2)
        units = [[cell.value for cell in iter] for iter in unit_iter][0]
        for k in range(4, ws.max_row + 1):
            # for k in range(301, 350 + 1):

            rows_iter = ws.iter_rows(min_col=ws.min_column, min_row=k, max_col=ws.max_column, max_row=k)
            row = [[cell.value for cell in iter] for iter in rows_iter][0]
            if row[title.index('Bomatec Bezeichnung')] is None:
                continue

            val = {}
            additional_files = ""
            for name in select:
                cell = row[title.index(name)]
                # print(name + ' :: ' + str(cell))
                if name == 'Bomatec Bezeichnung':
                    val[name] = cell
                    val['name'] = val[name]
                    val['component_name'] = cell + '_Bomatec'
                    additional_files = cell.replace('/', '_').replace(' ', '')
                    val['pdf_name'] = additional_files
                    val['png_name'] = additional_files
                    hclim_file = additional_files
                    val['info_displayname'] = cell
                elif name == 'Werkstoff':
                    val[name] = cell
                elif name in ['Temperaturstabil', 'Korrosionsstabil', 'isoptrop', 'anisotrop', 'gesintert', 'gegossen']:
                    val[name] = cell
                else:
                    property_value = MaterialBomatec.get_value(cell)
                    if property_value is None:
                        if name == "specific el. resistance":
                            if val['Werkstoff'] is 'Ferrite injection molded':
                                property_value = 1000000000
                            elif val['Werkstoff'] is 'SmCo injection molded':
                                property_value = None
                            elif val['Werkstoff'] is 'SmFeN injection molded':
                                property_value = None
                            elif val['Werkstoff'] is 'NdFeB injection molded':
                                property_value = None

                    if property_value is not None:
                        val[MaterialBomatec.get_name(name)] = {"variable": name,
                                                               "value": property_value,
                                                               "unit": MaterialBomatec.get_units(name, title, units),
                                                               "comment": ""
                                                               }

            val['Poisson ratio'] = {"variable": "nu", "value": "0.24", "unit": "[1]", "comment": ""}
            val['class'] = 'Permanent magnet'
            val['manufacturer'] = 'Bomatec'
            val['webpath'] = MaterialBomatec.get_webpath(val['Werkstoff'])
            val['ColorR'] = {"variable": "ColorR", "value": 255, "unit": "", "comment": ""}
            val['ColorG'] = {"variable": "ColorG", "value": 255, "unit": "", "comment": ""}
            val['ColorB'] = {"variable": "ColorB", "value": 0, "unit": "", "comment": ""}

            # Read JSON file for demagnetization
            json_file = os.path.join(self.path, 'resources', 'Hc_lim', additional_files + '.json')
            if os.path.exists(json_file):
                with open(json_file) as fh:
                    demag = json.loads(fh.read())

                Hcb = np.array([])
                for key in demag:
                    theta = float(key.replace('degC', ''))
                    Hcb = np.append(Hcb, np.array([theta, demag[key] * -1000]))

                Hcb = Hcb.reshape(len(demag.keys()), 2)
                idx = np.argsort(Hcb, axis=0)

                val['HvsTheta'] = {}
                val['HvsTheta']['Theta'] = {"variable": "Theta", "value": Hcb[idx[:, 0], 0].tolist(), "unit": "[degC]", "comment": ""}
                val['HvsTheta']['Hcb'] = {"variable": "Hcb", "value": Hcb[idx[:, 0], 1].tolist(), "unit": "[A/m]", "comment": ""}

            all_materials_dictionary[val['component_name']] = val

        n_mat = len(all_materials_dictionary.keys())

        for key in all_materials_dictionary.keys():
            material_dictionary = all_materials_dictionary[key]

            # print(material_dictionary)
            json_folder = str(os.path.join(self.path, 'resources'))
            with open(str(os.path.join(json_folder, material_dictionary['component_name'].replace('/', '_').replace(' ', '').replace('_Bomatec', '') + ".json")), "w") as f:
                json.dump(material_dictionary, f, indent=4, sort_keys=False)

    @staticmethod
    def get_webpath(path):
        magnet_types = ['AlNiCo', 'AlNiCo sintered', 'AlNiCo cast', 'Ferrite', 'Ferrite isotropic',
                        'Ferrite anisotropic', 'Ferrite injection molded', 'NdFeB', 'NdFeB bonded', 'NdFeB extruded',
                        'NdFeB injection molded', 'SmCo', 'SmCo injection molded', 'SmFeN injection molded']

        if path in magnet_types:
            webpath = 'Material.Metal.MagneticAlloy.{}'.format(path)
        elif path == 'Sm2Co17' or path == 'SmCo5':
            webpath = 'Material.Metal.MagneticAlloy.{}'.format('SmCo')
        else:
            raise Exception(path + " no considered magnet class.")

        return webpath

    @staticmethod
    def get_name(name):
        # 'Bomatec Bezeichnung'

        if name == 'Dichte ca':
            dict_name = 'Density'
        elif name == 'modulus of elasticity E':
            dict_name = 'Young modulus'
        elif name == 'flexural strength':
            dict_name = 'Tensile strength'
        elif name == 'vickers hardness HV':
            dict_name = 'Hardness'
        elif name == 'specific heat capacity':
            dict_name = 'Specific thermal capacity'
        elif name == 'thermal conductivity':
            dict_name = 'Thermal conductivity'
        elif name == 'specific el. resistance':
            dict_name = 'Resistivity'
        elif name == 'Br min':
            dict_name = 'Remanence flux density'
        elif name == 'permeability 20°C µr':
            dict_name = 'Relative permeability'
        elif name == 'temperature coefficient Br':
            dict_name = 'Temperature coefficient remanence flux density'
        elif name == 'T max':
            dict_name = 'Max temperature'
        else:
            dict_name = name

        return dict_name


    @staticmethod
    def get_value(cell):
        value = None

        if isinstance(cell, str) and '~' in cell:
            rng = cell.split('~')
            value = (float(rng[0].strip()) + float(rng[1].strip())) / 2

        elif isinstance(cell, str) and '-' in cell:
            rng = cell.split('-')
            try:
                value = (float(rng[0].strip()) + float(rng[1].strip())) / 2
            except:
                value = None

        elif isinstance(cell, str) and 'ca.' in cell:
            value = cell.replace('ca.', '')

        else:
            value = cell

        return value

    @staticmethod
    def get_units(name, title, units):
        unit = units[title.index(name)]
        print(unit)

        if unit == "[µΩ.m]":
            unit = "[uOhm.m]"
        elif unit == "[\u00b0C]":
            unit = "[degC]"
        elif unit == "[%/\u00b0C]":
            unit = "[%.K-1]"
        elif unit is None:
            unit = "[1]"
        elif name == "thermal conductivity" and unit == "[W/m.K]":
            unit = "[W/(m.K)]"

        return unit