# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 4767 $
$LastChangedDate:: 2021-09-07 11:29:07 +0200#$

This file is part of SyMSpace.

Copyright (c) 2021, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import glob
import json
import tempfile
from symspace.utils.logger import LoggerClass
from symspace.osutils.pathutils import Paths
from symspace.wrapper.symwrapper import SymProject, SymContainer, SymParameter, SymField
from symspace.osutils.multiprocessing import ExternalMultiprocessing
from componentspace.utils.scripts import run_external_template_check


class ComponentTemplateCheck(LoggerClass):
    """Compare Component-SyMSpace Projects to a Template Project

       And add missing fields etc. update comments and descriptions
    """

    def __init__(self, template_mop: str = None, loglevel: str = 'WARNING', log_file: str = None):
        super().__init__(loglevel=loglevel, log_file=log_file)
        self.__log_file = log_file
        self.__cs_root = Paths.getComponentSpaceRoot()
        if template_mop is None:
            self.__template_mop = os.path.join(self.__cs_root, '_template', 'Template.mop')

    def getTemplateParamStructure(self) -> dict:
        template_prj = SymProject.load(self.__template_mop)
        template_param_dict = dict()
        self.__getIdentifierR(template_prj, template_param_dict)
        template_param_dict.pop('Component.Project.ReplaceThis')
        template_prj.close()
        return template_param_dict

    def writeTemplateParamStructure(self) -> str:
        template_param_dict = self.getTemplateParamStructure()
        temp_dir = tempfile.mkdtemp()
        json_file = os.path.join(temp_dir, 'template.json')

        with open(json_file, 'w') as fh:
            json.dump(template_param_dict, fh, indent=4, sort_keys=True)
        return json_file

    def getMopFiles(self, start_path: str = None) -> list:
        if start_path is None:
            start_path = self.__cs_root
        mop_files = glob.glob(os.path.join(start_path, '**', '*.mop'), recursive=True)
        return mop_files

    def getTemplateMopFiles(self, start_path: str = None) -> list:
        if start_path is None:
            start_path = self.__cs_root
        mop_files = glob.glob(os.path.join(start_path, '**', 'Template.mop'), recursive=True)
        return mop_files

    def start(self, mop_files: list, json_file: str,
              dry_run: bool = False, check_comment: bool = False, check_description: bool = False,
              max_cpu_percent: float = 75.0, max_ram_percent: float = 75.0, timeout: float = 60.0):
        xmp = ExternalMultiprocessing(max_cpu_percent=max_cpu_percent, max_ram_percent=max_ram_percent, timeout=timeout)
        n_files = len(mop_files)
        script_py = run_external_template_check.__file__

        for index, mop_file in enumerate(mop_files):
            self.logger.info('Check file #{:d} / {:d}: {}'.format(index + 1, n_files, mop_file))

            cmd_args = [Paths.getCSyMSpaceExe(), script_py, mop_file, json_file,
                        int(dry_run), int(check_comment), int(check_description)]
            xmp.startProcess(cmd_args, wait_to_start=False, wait_to_finish=False)
            xmp.wait()

        '''wait until all processes are done'''
        xmp.waitForCompletion()

        self.logger.info('___Summary___')
        xmp.refresh()

        for index, process in enumerate(xmp.processes):
            pid = process.pid
            pid_status = xmp.pid_status[pid]

            if pid_status['Error'] and pid_status['stoppedByTimeout']:
                status = 'Failed, Timeout'
                log_handle = self.logger.error
            elif pid_status['Error']:
                status = 'Failed'
                log_handle = self.logger.error
            elif pid_status['stoppedByTimeout']:
                status = 'Timeout'
                log_handle = self.logger.error
            else:
                status = 'Successful'
                log_handle = self.logger.info

            log_handle('Check status  #{:d} / {:d}: {} '.format(index + 1, n_files, status))

            self.logPIDStatus(pid_status['StdOutLog'], self.logger.info)
            self.logPIDStatus(pid_status['StdErrLog'], self.logger.error)

        if os.path.isfile(self.__log_file):
            print('Logfile: {}'.format(self.__log_file))

    @staticmethod
    def logPIDStatus(stdlog: bytes, log_handle):
        try:
            msg = stdlog.decode('utf-8')
        except Exception as ex:
            msg = ''
        if msg != '' and msg != '\n':
            log_handle(msg)

    @classmethod
    def __getIdentifierR(cls, param: SymParameter, d: dict):
        param_id = param.getIdentifier()
        if isinstance(param, SymContainer):
            d[param_id] = dict(Description=param.getDescription(), Comment=param.getComment())
            for sub_param in param.getParameters():
                cls.__getIdentifierR(sub_param, d)
        elif isinstance(param, SymField):
            d[param_id] = dict(Description=param.getDescription(), Comment=param.getComment())
        else:
            raise NotImplementedError('Currently only SymFields and SymContainer are considered, '
                                      'got {} instead'.format(type(param)))


if __name__ == '__main__':
    _start_path = None
    # _start_path = os.path.join(Paths.getComponentSpaceRoot(), 'Material', 'Metal')
    _log_file = os.path.join(tempfile.mkdtemp(), 'template_check.log')

    _ctc = ComponentTemplateCheck(loglevel='INFO', log_file=_log_file)

    # _template_files = _ctc.getTemplateMopFiles()
    # for _file in _template_files:
    #     print(_file)

    _mop_files = _ctc.getMopFiles(_start_path)
    _json_file = _ctc.writeTemplateParamStructure()

    _ctc.start(_mop_files, _json_file, dry_run=False, check_comment=False, check_description=False,
               max_cpu_percent=75., max_ram_percent=75.)
